config_not_found = "Terality configuration file not found."  # pylint: disable=invalid-name


# pylint: disable=invalid-name
config_helper = """
- If you don't have a Terality account yet, go to https://terality.com/ to sign up!
- If you have a Terality account and received your credentials, call `terality.configure(user_id, user_password)` to set up your configuration.
- If you have moved the configuration somewhere else, specify its location using the `TERALITY_HOME` environment variable.
"""
